/*******************************************************************
  uLan Communication - uL_DRV - multiplatform uLan driver

  ul_linux.c	- Linux kernel and module specific part

  (C) Copyright 1996-2004 by Pavel Pisa - project originator
        http://cmp.felk.cvut.cz/~pisa
  (C) Copyright 1996-2004 PiKRON Ltd.
        http://www.pikron.com
  (C) Copyright 2002-2004 Petr Smolik
  

  The uLan driver project can be used and distributed 
  in compliance with any of next licenses
   - GPL - GNU Public License
     See file COPYING for details.
   - LGPL - Lesser GNU Public License
   - MPL - Mozilla Public License
   - and other licenses added by project originator

  Code can be modified and re-distributed under any combination
  of the above listed licenses. If contributor does not agree with
  some of the licenses, he/she can delete appropriate line.
  WARNING: if you delete all lines, you are not allowed to
  distribute code or sources in any form.
 *******************************************************************/

/*******************************************************************/
/* Linux kernel and module specific part */ 

MODULE_SUPPORTED_DEVICE("ul_drv");
MODULE_AUTHOR("Pavel Pisa <pisa@cmp.felk.cvut.cz>");
MODULE_DESCRIPTION("\
uLan communication module ( 9 bit RS-485 protocol )\n\
for communication with embedded controllers.\n\
It uses i82510 addon card or RS-232 dongle");
MODULE_LICENSE("GPL");
EXPORT_NO_SYMBOLS;

static int ulan_major_dev=ULAN_MAJOR;	/* uLan major device number */

#define UL_MINORS 9
ul_drv *ul_drv_arr[UL_MINORS]={[0 ... UL_MINORS-1]=NULL};
char *chip[UL_MINORS]={NULL,};
#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)
char *slot[UL_MINORS]={NULL,};
#endif
int my_adr[UL_MINORS]={[0 ... UL_MINORS-1]=2};
int baud[UL_MINORS]={[0 ... UL_MINORS-1]=19200};
int irq[UL_MINORS]={0,};
int port[UL_MINORS]={0x3e8 
	#if (UL_MINORS>1)
		,[1 ... UL_MINORS-1]=0
	#endif
		};
int debug=-1;		/* debug flags */
int ul_usb_msg_inpr=0;	/* number of concurrent messages sent over USB */

MODULE_PARM_DESC(chip,"chip type: auto, 82510, 16450, 16950-pci");
MODULE_PARM(chip,"1-" __MODULE_STRING(UL_MINORS) "s");
MODULE_PARM_DESC(my_adr,"uLan address of defined interface");
MODULE_PARM(my_adr,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(baud,"define baudrates");
MODULE_PARM(baud,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(irq,"define irqs");
MODULE_PARM(irq,"1-" __MODULE_STRING(UL_MINORS) "i");
MODULE_PARM_DESC(port,"define ports");
MODULE_PARM(port,"1-" __MODULE_STRING(UL_MINORS) "i");
#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)
MODULE_PARM_DESC(slot,"slot type (pci) and specification");
MODULE_PARM(slot,"1-" __MODULE_STRING(UL_MINORS) "s");
#endif
MODULE_PARM_DESC(debug,"debug flags (FATAL,CHIO,IRQ,MSG,FAILS,"\
		       "SEQ,PORTS,FILE,FILT)");
MODULE_PARM(debug,"i");
MODULE_PARM_DESC(ul_usb_msg_inpr,"used length of Tx fifo in conventor");
MODULE_PARM(ul_usb_msg_inpr,"i");
MODULE_PARM_DESC(ulbuffer,"size of per interface buffer");
MODULE_PARM(ulbuffer,"i");

#ifdef UL_WITH_PCI

static int /*__devinit*/ ulan_pci_init_one (struct pci_dev *dev,
                                   const struct pci_device_id *ent);

static void /*__devexit*/ ulan_pci_remove_one (struct pci_dev *dev);

MODULE_DEVICE_TABLE(pci, ulan_pci_tbl);

static struct pci_driver ulan_pci_driver = {
	name:		"ul_drv",
	id_table:	ulan_pci_tbl,
	probe:		ulan_pci_init_one,
	remove:		ulan_pci_remove_one,
};
#endif /*UL_WITH_PCI*/

#ifdef UL_WITH_USB

#if (LINUX_VERSION_CODE < VERSION(2,5,41))
static void *ul_usb_probe(struct usb_device *dev, unsigned int ifnum, const struct usb_device_id *id);
static void ul_usb_disconnect(struct usb_device *dev, void *ptr);
#else /* 2.5.41 */
static int ul_usb_probe(struct usb_interface *intf, const struct usb_device_id *id);
static void ul_usb_disconnect(struct usb_interface *intf);
#endif /* 2.5.41 */

MODULE_DEVICE_TABLE (usb, ulan_usb_tbl);

#ifdef CONFIG_USB_DYNAMIC_MINORS
/* 
 * if the user wants to use dynamic minor numbers, then we can have up to 256
 * devices
 */
#define UL_USB_MINOR_BASE	0
#define MAX_DEVICES		256
#else
/* Get a minor range for your devices from the usb maintainer */
#define UL_USB_MINOR_BASE	200

/* we can have up to this number of device plugged in at once */
#define MAX_DEVICES		16
#endif

/* usb specific object needed to register this driver with the usb subsystem */
static struct usb_driver ul_usb_driver = {
	name:		"ulan_usb",
	probe:		ul_usb_probe,
	disconnect:	ul_usb_disconnect,
	/*fops:		&ul_usb_fops,*/
	/* num_minors:	MAX_DEVICES,*/
     #if (LINUX_VERSION_CODE < VERSION(2,5,8))
	minor:		UL_USB_MINOR_BASE,
     #endif
	id_table:	ulan_usb_tbl,
};

#endif /*UL_WITH_USB*/


static int ulan_open(struct inode *inode, struct file *file);
static CLOSERET ulan_close(struct inode *inode, struct file *file);
static int ulan_ioctl(struct inode *inode, struct file *file, 
			unsigned int cmd, unsigned long arg);
static RWRET ulan_write(WRITE_PARAMETERS);
static RWRET ulan_read(READ_PARAMETERS);
static loff_t ulan_lseek(LSEEK_PARAMETERS);
static unsigned int ulan_poll(struct file *, poll_table *);

/* static struct file_operations ulan_fops */
KC_CHRDEV_FOPS_BEG(ulan_fops)
	KC_FOPS_LSEEK(ulan_lseek)/* lseek              */
	read:ulan_read,		/* read                */
	write:ulan_write,	/* write               */
	readdir:NULL,		/* readdir             */
	poll:ulan_poll,		/* poll                */
	ioctl:ulan_ioctl,	/* ioctl               */
	mmap:NULL,		/* mmap                */
	open:ulan_open,		/* open                */
	KC_FOPS_FLUSH(NULL)	/* flush               */
	KC_FOPS_RELEASE(ulan_close)/* close/release    */
	fsync:NULL,		/* fsync               */
	fasync:NULL,		/* fasync              */
	NULL,			/* check_media_change  */
	NULL,			/* revalidate          */
KC_CHRDEV_FOPS_END;


/* open ulan device */
static int ulan_open(struct inode *inode, struct file *file)
{ 
  ul_drv *udrv;
  ul_opdata *opdata;
  unsigned int minor = kc_dev2minor(inode->i_rdev);
  #ifdef UL_WITH_DEVFS
  if(minor==-1){
    udrv=file->private_data;
  }else
  #endif /* UL_WITH_DEVFS */
  {
    if ((minor >= UL_MINORS)||(minor<0)) return -ENODEV;
    udrv=ul_drv_arr[minor];
  }
  if (!udrv) return -ENODEV;
  if(udrv->magic!=UL_DRV_MAGIC)
  {
    printk(KERN_CRIT "ulan_open: Wrong uLan MAGIC number!!!\n");
    return -ENODEV;
  }

  kc_MOD_INC_USE_COUNT;
  if(!(file->private_data=MALLOC(sizeof(ul_opdata))))
  { kc_MOD_DEC_USE_COUNT;
    return(-ENOMEM);
  };
  opdata=(ul_opdata*)file->private_data;
  opdata->file=file;
  opdata->magic=ULOP_MAGIC;
  opdata->message=NULL;
  opdata->udrv=udrv;
  init_waitqueue_head(&opdata->wqrec);
  opdata->opprew=NULL;
  opdata->opnext=NULL;
  opdata->recchain=NULL;
  opdata->filtchain=NULL;
  { /* add us onto list of clients of udrv */
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    opptr=udrv->operators;
    if(opptr) {opptr->opprew=opdata;opdata->opnext=opptr;};
    UL_MB();
    udrv->operators=opdata;
    UL_DRV_UNLOCK;
  };

  return 0;
};

static CLOSERET ulan_close(struct inode *inode, struct file *file)
{
  ul_opchain *opmember;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_close : BAD opdata magic !!!");

  if (opdata->message) ulan_freemsg(opdata);
  { /* delete us from list of clients of udrv */
    ul_drv *udrv;
    ul_opdata *opptr;
    UL_DRV_LOCK_FINI
    UL_DRV_LOCK;
    if((udrv=opdata->udrv)!=NULL){
      if((opptr=opdata->opnext)) opptr->opprew=opdata->opprew;
      if((opptr=opdata->opprew)) opptr->opnext=opdata->opnext;
      else udrv->operators=opdata->opnext;
    }
    UL_DRV_UNLOCK;
    if(udrv)
      if(udrv->irq>0) ul_synchronize_irq(udrv->irq);
    while((udrv=opdata->udrv)!=NULL){
      if(!uld_atomic_test_dfl(udrv,IN_BOTTOM))
        break;
      kc_yield();
    }
  }
  schedule();
  while((opmember=opdata->recchain))
  {
    del_from_opchain(&opdata->recchain,opmember);
    if(opmember->message) ul_dec_ref_cnt(opmember->message);
    FREE(opmember);
  };
  while((opmember=opdata->filtchain))
  {
    del_from_opchain(&opdata->filtchain,opmember);
    FREE(opmember);
  };
  FREE(file->private_data);
  kc_MOD_DEC_USE_COUNT;
  return (CLOSERET)0;
};

static int ulan_ioctl(struct inode *inode, struct file *file, 
			unsigned int cmd, unsigned long arg)
{
  ul_msginfo msginfo;
  int ret;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_ioctl : BAD opdata magic !!!");

  switch (cmd) {
   case UL_DRV_VER :
	return UL_DRV_VERCODE;
   case UL_NEWMSG :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_newmsg(opdata,&msginfo);
   case UL_TAILMSG :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_tailmsg(opdata,&msginfo);
   case UL_FREEMSG :
  	return ulan_freemsg(opdata);
   case UL_ACCEPTMSG :  	
	if(kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo)))
	   return -EFAULT;
	ret=ulan_acceptmsg(opdata,&msginfo);
	kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo));
  	return  ret;
   case UL_ACTAILMSG :  	
	if(kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo)))
	   return -EFAULT;
	ret=ulan_actailmsg(opdata,&msginfo);
	kc_copy_to_user((void *)arg,&msginfo,sizeof(msginfo));
  	return  ret;
   case UL_ADDFILT :
	if(kc_copy_from_user(&msginfo,(void *)arg,sizeof(msginfo)))
	   return -EFAULT;
  	return ulan_addfilt(opdata,&msginfo);
   case UL_ABORTMSG :
  	return ulan_abortmsg(opdata);
   case UL_REWMSG :
  	return ulan_rewmsg(opdata);
   case UL_KLOGBLL :
        if(!opdata->udrv) return -ENODEV;
	printudrvbll(opdata->udrv);
    #ifdef ENABLE_UL_MEM_CHECK
	printudrvoperators(opdata->udrv);
    #endif /* ENABLE_UL_MEM_CHECK */
	return 0;
   case UL_STROKE :
        if(!opdata->udrv) return -ENODEV;
	ulan_stroke(opdata->udrv);
	return 0;
   case UL_DEBFLG :
	uld_debug_flg=arg;
	return 0;
   case UL_INEPOLL :
        return ulan_inepoll(opdata);
   case UL_HWTEST :
        if(current->euid!=0) return -EPERM;
	return ulan_hwtest(opdata,arg);
  };


  return -EINVAL;
};

/* common parameters (struct file *file, const char *buf, 
                      size_t count [, loff_t *ppos]) */
static RWRET ulan_write(WRITE_PARAMETERS)
{
  int cn;
  int len;
  ul_mem_blk *blk;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_write : BAD opdata magic !!!");

  if(!opdata->udrv) return -ENODEV;
  if(!opdata->message) return -ENOMSG;

  cn=count;
  while(cn>0)
  {
    while(!ul_di_adjust(&opdata->data))
    {
      if(!(blk=ul_alloc_blk(opdata->udrv))) 
	{count-=cn;cn=0;break;};
      memset(UL_BLK_NDATA(blk),0,UL_BLK_SIZE);
      opdata->data.blk->next=blk;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    kc_copy_from_user(ul_di_byte(&opdata->data),buf,len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  if(opdata->data.pos>UL_BLK_HEAD(opdata->data.head_blk).len)
    UL_BLK_HEAD(opdata->data.head_blk).len=opdata->data.pos;

  return count-cn;
};

/* common parameters (struct file *file, char *buf, 
                      size_t count [, loff_t *ppos]) */
static RWRET ulan_read(READ_PARAMETERS)
{
  int cn;
  int len;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_read : BAD opdata magic !!!");

  if(!opdata->udrv) return -ENODEV;
  if(!opdata->message) return -ENOMSG;

  if(opdata->data.pos+count>UL_BLK_HEAD(opdata->data.head_blk).len)
   count=UL_BLK_HEAD(opdata->data.head_blk).len-opdata->data.pos;

  cn=count;
  while(cn>0)
  {
    if(!ul_di_adjust(&opdata->data))
    {
      while(cn--) kc_put_user(0,buf++);
      break;
    };
    len=ul_di_atonce(&opdata->data);
    if(len>cn) len=cn;
    kc_copy_to_user(buf,ul_di_byte(&opdata->data),len);
    ul_di_add(&opdata->data,len);
    buf+=len;
    cn-=len;
  };

  return count-cn;
};

/* common parameters (struct file *file, loff_t pos, int whence) */
static loff_t ulan_lseek(LSEEK_PARAMETERS)
{
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_lseek : BAD opdata magic !!!");

  if(!opdata->message) return -ENOMSG;

  return ul_di_seek(&opdata->data,pos,whence);
};

static unsigned int ulan_poll(struct file *file, poll_table *wait)
{
  /* POLLIN | POLLOUT | POLLERR | POLLHUP | POLLRDNORM | POLLWRNORM */
  unsigned int mask = 0;
  ul_opdata *opdata=(ul_opdata*)file->private_data;
  if(opdata->magic!=ULOP_MAGIC) panic("ulan_poll : BAD opdata magic !!!");

  kc_poll_wait(file, &opdata->wqrec, wait);

  if (opdata->recchain)
    mask |= (POLLIN | POLLRDNORM);

  if (!opdata->udrv)
    mask |= (POLLIN | POLLOUT | POLLERR);

  return mask;
};

#if defined(UL_WITH_PCI) || defined(UL_WITH_USB)

static int /*__devinit*/ 
 ulan_init_find_minor(char *abus, char *aslot, char *asubdev,
                      int *pminor, int *ppar)
{
  int i;
  int abus_len;
  int aslot_len;
  int sub_offs;
  int minor;
  int par;
  int match=0;
  char *islot;
  abus_len=strlen(abus);
  aslot_len=strlen(aslot);
  sub_offs=abus_len+aslot_len;
  *pminor=*ppar=minor=par=-1;
  for(i=0;i<UL_MINORS;i++){
    islot=slot[i];
    if(islot)
      if(!*islot) islot=NULL;
    switch (match){
      case 0: 
        if(!ul_drv_arr[i]&&!islot){
	  minor=i;match=1;
	  if(par==-1) par=i;
	}
      case 1:
        if(!islot) continue;
        if(!strcmp(islot,abus)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=2;}
	}
      case 2:
        if(!islot||!aslot) continue;
        if(!strncmp(islot,abus,abus_len)&&
	   !strcmp(islot+abus_len,aslot)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=3;}
        }
      case 3:
        if(!islot||!aslot||!asubdev) continue;
        if(!strncmp(islot,abus,abus_len)&&
	   !strncmp(islot+abus_len,aslot,aslot_len)&&
	   (islot[sub_offs]=='.')&&!strcmp(islot+sub_offs+1,asubdev)){
	  par=i;
	  if(!ul_drv_arr[i]){ minor=i; match=4;}
        }
    }
  }
  *pminor=minor;*ppar=par;
  return match;
}

#endif /*defined(UL_WITH_PCI)||defined(UL_WITH_USB)*/


/*** initialize kernel module ***/
int init_module(void)
{
 #ifdef UL_WITH_DEVFS
  kc_devfs_handle_t devfs_handle;
  char devname[32];
 #endif /* UL_WITH_DEVFS */
  ul_drv* udrv;
  int ret;
  int minor;

  printk(KERN_INFO "uLan v" UL_DRV_VERSION " init_module\n");
  printk(KERN_INFO "Compiled at %s %s\n",__DATE__, __TIME__);

  if(debug!=-1){
    uld_debug_flg=debug;
  }

  ret=0;
 #ifdef UL_WITH_UARTS
  for(minor=0;minor<UL_MINORS;minor++)
  {
    if(port[minor]||chip[minor])
    { /* get new ul_drv (port, irq, baud, my_adr); */
      udrv=ul_drv_new(port[minor],irq[minor],
   		      baud[minor],my_adr[minor],chip[minor]);
      ul_drv_arr[minor]=udrv;
      if(udrv){
	#ifdef UL_WITH_DEVFS
	  sprintf (devname, "ulan%d", minor);
          devfs_handle=kc_devfs_new_cdev(NULL, MKDEV(ulan_major_dev, minor), 
			S_IFCHR | S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, 
			&ulan_fops, udrv, devname);
	  udrv->devfs_handle=devfs_handle;
	#endif /* UL_WITH_DEVFS */
	ret++;
      };
    };
  };
 #endif /*UL_WITH_UARTS*/

 #ifdef UL_WITH_PCI
  { int rc;
    rc=pci_register_driver (&ulan_pci_driver);
    ret+=rc;
  }
 #endif  /* UL_WITH_PCI */

 #ifdef UL_WITH_USB
  /* register this driver with the USB subsystem */
  { int rc;
    rc = usb_register(&ul_usb_driver);
    if (rc < 0) {
      printk(KERN_CRIT "usb_register failed for the "__FILE__" driver. Error number %d",
	 rc);
    } else {
      ret++;
    }
  }
 #endif  /* UL_WITH_USB */

  if(!ret)
  { printk(KERN_ERR "uLan module_init : no hardware initialized, aborting !!!\n");
    #ifdef UL_WITH_USB
     usb_deregister(&ul_usb_driver);
    #endif  /* UL_WITH_USB */
    #ifdef UL_WITH_PCI
    pci_unregister_driver (&ulan_pci_driver);
    #endif  /* UL_WITH_PCI */
    return -ENODEV;
  };

 #ifdef UL_WITH_DEVFS
  ret = devfs_register_chrdev(ulan_major_dev, "ulan", &ulan_fops);
 #else /* UL_WITH_DEVFS */
  ret = register_chrdev(ulan_major_dev, "ulan", &ulan_fops);
 #endif /* UL_WITH_DEVFS */
  printk(KERN_INFO "uLan init : called register dev %d and got %d\n",
			 ulan_major_dev,ret);
  if (ret < 0) {
    printk(KERN_ERR "uLan init : unable to grab device #\n");
    ulan_major_dev=0;
    #ifdef UL_WITH_PCI
    pci_unregister_driver (&ulan_pci_driver);
    #endif  /* UL_WITH_PCI */
    for(minor=0;minor<UL_MINORS;minor++)
    { udrv=ul_drv_arr[minor];
      if(udrv)
      { ul_drv_arr[minor]=NULL;
	#ifdef UL_WITH_DEVFS
	if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
	#endif /* UL_WITH_DEVFS */
	ul_drv_free(udrv);
      };
    };
    return ret;
  }
  if(!ulan_major_dev) ulan_major_dev=ret;
  printk(KERN_INFO "uLan init : driver succesfully registered as major %d\n",
	          ulan_major_dev);

  #ifdef UL_INT_TST
  udrv=ul_drv_arr[0];
  if(udrv)
  { printk("uLan : geninfoblk\n");
    geninfoblk(udrv);
    printk("uLan : gentestblk\n");
    gentestblk(udrv);
    printk("uLan : uld_timeout\n");
    uld_timeout(udrv);
  };
  #endif /* UL_INT_TST */

  return 0;
}

/*** cleanup kernel module ***/
void cleanup_module(void)
{
  ul_drv* udrv;
  int i;
  int ret;

  if (ulan_major_dev != 0) {
   #ifdef UL_WITH_DEVFS
    ret = devfs_unregister_chrdev(ulan_major_dev, "ulan");
   #else /* UL_WITH_DEVFS */
    ret = unregister_chrdev(ulan_major_dev, "ulan");
   #endif /* UL_WITH_DEVFS */
    if(ret>=0)
      printk(KERN_INFO "uLan cleanup: Unregister  O.K.\n");
    else
      printk(KERN_CRIT "uLan cleanup: Unregister failed\n");
  };

  udrv=ul_drv_arr[0];
  if(udrv)
  { printk("uLan : printudrvbll\n");
    printudrvbll(udrv);
   #ifdef UL_WITH_FRAME_FSM
    printk("uLan : max cycle cnt  %d\n",ul_max_cycle_cnt);
    printk("uLan : max cycle time %d\n",ul_max_cycle_time);
   #endif /*UL_WITH_FRAME_FSM*/
  }

 #ifdef UL_WITH_USB
  usb_deregister(&ul_usb_driver);
 #endif  /* UL_WITH_USB */
 #ifdef UL_WITH_PCI
  pci_unregister_driver (&ulan_pci_driver);
 #endif  /* UL_WITH_PCI */
  for(i=0;i<UL_MINORS;i++)
  { udrv=ul_drv_arr[i];
    ul_drv_arr[i]=NULL;
    if(udrv){
     #ifdef UL_WITH_DEVFS
     if(udrv->devfs_handle) kc_devfs_delete(udrv->devfs_handle);
     #endif /* UL_WITH_DEVFS */
     ul_drv_free(udrv);
    }
  }
 #ifdef ENABLE_UL_MEM_CHECK
  UL_PRINTF("MEM_CHECK: malloc-free=%d\n",
		 (int)atomic_read(&ul_mem_check_counter));
 #endif /* ENABLE_UL_MEM_CHECK */
}

#ifdef ENABLE_UL_MEM_CHECK
void * ul_mem_check_malloc(size_t size)
{void *ptr;
 ptr=kmalloc(size,GFP_KERNEL);
 if(ptr) atomic_inc(&ul_mem_check_counter);
 return ptr;
}

void ul_mem_check_free(void *ptr)
{
  if(ptr) atomic_dec(&ul_mem_check_counter);
  kfree(ptr);
}
#endif /* ENABLE_UL_MEM_CHECK */
